package hashing;

import java.util.LinkedList;

public class LinearProbingHashSet {

    // The example is to learn about linear probing, so I'm not actually bothering
    // to write anything for resizing. But... resizing would be helpful to have!
    private static final int INITIAL_SIZE = 10;

    // Need special values to indicate why cells might be empty.
    // WARNING: as a result, we can only hash positive values :(
    private static final int EMPTY_SINCE_START = -1;
    private static final int EMPTY_SINCE_REMOVAL = -2;


    private int[] table;
    private int M;
    private int n;

    public LinearProbingHashSet() {
        table = new int[INITIAL_SIZE];
        M = table.length;
        n = 0;

        // add the buckets!
        for (int i = 0; i < table.length; i++) {
            table[i] = EMPTY_SINCE_START;
        }

    }

    public boolean add(int elem) {
        // Hash function determines initial bucket
        int bucket = h(elem);
        int bucketsProbed = 0;
        while (bucketsProbed < M) {
            // Insert item in next empty bucket
            if (table[bucket] == EMPTY_SINCE_REMOVAL || table[bucket] == EMPTY_SINCE_START) {
                table[bucket] = elem;
                return true;
            }
            // Increment bucket index
            bucket = (bucket + 1) % M;

            // Increment number of buckets probed
            bucketsProbed++;
        }

        return false;
    }

    public boolean remove(int elem) {
        // Hash function determines initial bucket
        int bucket = h(elem);
        int bucketsProbed = 0;

        while (table[bucket] != EMPTY_SINCE_START && bucketsProbed < M) {
            if (table[bucket] == elem) {
                table[bucket] = EMPTY_SINCE_REMOVAL;
                return true;
            }

            // Increment bucket index
            bucket = (bucket + 1) % M;

            // Increment number of buckets probed
            bucketsProbed++;
        }
        return false;
    }

    public boolean contains(int key) {
        // Hash function determines initial bucket
        int bucket = h(key);
        int bucketsProbed = 0;

        while (table[bucket] != EMPTY_SINCE_START && bucketsProbed < M) {

            if (table[bucket] == key) {
                return true;
            }

            // Increment bucket index
            bucket = (bucket + 1) % M;

                    // Increment number of buckets probed
            bucketsProbed++;
        }

        return false;  // Item not found
    }

    public int h(Object o) {
        // need to use o.hashCode()

        // need to take the hash code and put it
        // in range [0, M)
        return o.hashCode() % M;
    }
}
