package hashing;

import java.util.Arrays;

public class NoOverflowBucketHashSet<E> {

    // The example is to learn about buckets, so I'm not actually bothering
    // to write anything for resizing. But... resizing would be helpful to have!
    private static final int INITIAL_SIZE = 32;
    private static final int NUM_BUCKETS = 8;


    private Object[] table;
    private int numSlots;
    private int numBuckets;
    private int n;
    private int slotsPerBucket;


    public NoOverflowBucketHashSet() {
        table = new Object[INITIAL_SIZE];
        numSlots = INITIAL_SIZE;
        numBuckets = NUM_BUCKETS;
        slotsPerBucket = numSlots / numBuckets;
        n = 0;

        // add the buckets!
        Arrays.fill(table, null);

    }

    public boolean add(E elem) {
        // Hash function determines initial bucket
        int bucket = h(elem);
        for (int slot = bucket * slotsPerBucket;
             slot < (bucket + 1) * slotsPerBucket;
             slot++) {
            // Insert item in next empty bucket
            if (table[slot] == null) {
                table[slot] = elem;
                return true;
            }
        }

        return false;
    }

    public boolean remove(Object elem) {
        // Hash function determines initial bucket
        int bucket = h(elem);
        int bucketStart = bucket * slotsPerBucket;
        for (int slot = bucketStart;
             slot < bucketStart + slotsPerBucket && table[slot] != null;
             slot++) {
            // Insert item in next empty bucket
            if (table[slot].equals(elem)) {
                table[slot] = elem;
                int numSlotsAfter = bucketStart + slotsPerBucket - slot - 1;
                System.arraycopy(table, slot + 1, table, slot, numSlotsAfter);
                return true;
            }
        }
        return false;
    }

    public boolean contains(E elem) {
        // Hash function determines initial bucket
        int bucket = h(elem);
        for (int slot = bucket * slotsPerBucket;
             slot < (bucket + 1) * slotsPerBucket;
             slot++) {
            // Insert item in next empty bucket
            if (table[slot] == elem) {
                return true;
            }
        }

        return false;
    }

    public int h(Object o) {
        // need to use o.hashCode()

        // need to take the hash code and put it
        // in range [0, M)

        System.out.println(o);
        System.out.println(o.hashCode() % numBuckets);
        return o.hashCode() % numBuckets;
    }

    @Override
    public String toString() {
        return "OverflowBucketHashSet{" +
                "table=" + Arrays.toString(table) +
                '}';
    }

    public static void main(String[] args) {
        NoOverflowBucketHashSet<Integer> hs = new NoOverflowBucketHashSet<>();
        for (int i : new int[] {0, 40, 44, 48, 157, 313, 930, 207, 979, 100, 977}) {
            hs.add(i);
        }
        System.out.println(hs);

        hs.remove(0);
        System.out.println(hs);

    }
}
