package hashing;

import com.sun.source.tree.Tree;

import java.util.*;

public class OpenHashSet<E> {

    private static final int INITIAL_SIZE = 10;
    private static final int MAX_LIST_SIZE = 2;
    private static final double CAPACITY = 0.75;


    // Collection<E>[]
    // for small buckets, List<E>
    // for larger buckets above MAX_LIST_SIZE, Tree<E>
    private Object[] table;
    private int M;
    private int n;

    public OpenHashSet() {
        table = new Object[INITIAL_SIZE];
        M = table.length;
        n = 0;

        // add the buckets!
        for (int i = 0; i < table.length; i++) {
            table[i] = new LinkedList<E>();
        }

    }

    public int h(Object o) {
        // need to use o.hashCode()

        // need to take the hash code and put it
        // in range [0, M)
        return o.hashCode() % M;
    }

    // guaranteed O(lg n) time in worst case
    public boolean contains(Object o) {
        // 1. get the index of the array where the object would live
        //    by calling h
        // 2. go get the list (or tree!) at that position in the array
        // 3. iterate through the collection to see if o is in there
        //    if it's in there, return true
        //    if it's not in there, return false

        int bucketIdx = h(o);
        Collection<E> bucket = (Collection<E>) table[bucketIdx];
        return bucket.contains(o);
    }

    public boolean add(E elem) {
        // refuse null elements
        if (elem == null) {
            return false;
        }

        // don't add duplicates!
        if (contains(elem)) {
            return false;
        }

        int bucketIdx = h(elem);
        Collection<E> bucket = (Collection<E>) table[bucketIdx];

        if (bucket.size() == MAX_LIST_SIZE) {
            // turn this bucket into a TreeSet instead of the LinkedList it currently is
            // to guarantee O(lg n) runtime for search/insert/remove operation

            // only changes the variable, not the reference in the array itself
            bucket = new TreeSet<>(bucket);
            table[bucketIdx] = bucket;
        }
        return bucket.add(elem);
    }

    private boolean addElemToBucketAtIdx(int bucketIdx, E elem) {
        return true;

    }

    public boolean remove(Object o) {
        return false;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < M; i++) {
            sb.append(i + ": " + table[i].getClass() + " of " + table[i] + "\n");
        }
        return sb.toString();
    }

    public static void main(String[] args) {
        OpenHashSet<Integer> ohs = new OpenHashSet<>();
        for (int i : new int[] {157, 313, 930, 207, 979, 100, 977}) {
            ohs.add(i);
        }
        System.out.println(ohs);
    }
}
